
import logging
logger = logging.getLogger(__name__)
logger.warning("dank_mids.executor module has been deprecated and will be removed eventually.")
logger.warning("you can now import what you need from a_sync.primitives module https://github.com/BobTheBuidler/ez-a-sync")
from a_sync.primitives.executor import *